@@ -462,6 +462,11 @@ enum iscsi_task_mgmt_funcs {
  *
  * The callback will NOT be invoked if the session is explicitely torn down
  * through a call to iscsi_disconnect() or iscsi_destroy_context().
+ *
+ * abort_task will also cancel the scsi task. The callback for the scsi task will be invoked with
+ *            SCSI_STATUS_CANCELLED
+ * abort_task_set, lun_reset, target_warn_reset, target_cold_reset will cancel all tasks. The callback for
+ *            all tasks will be invoked with SCSI_STATUS_CANCELLED 
  */
 EXTERN int
 iscsi_task_mgmt_async(struct iscsi_context *iscsi,
@@ -574,11 +579,25 @@ iscsi_readcapacity10_task(struct iscsi_context *iscsi, int lun, int lba,
 			       int pmi, iscsi_command_cb cb,
 			       void *private_data);
 EXTERN struct scsi_task *
+iscsi_readcapacity16_task(struct iscsi_context *iscsi, int lun,
+			       iscsi_command_cb cb,
+			       void *private_data);
+EXTERN struct scsi_task *
 iscsi_synchronizecache10_task(struct iscsi_context *iscsi, int lun,
 				   int lba, int num_blocks, int syncnv,
 				   int immed, iscsi_command_cb cb,
 				   void *private_data);
 EXTERN struct scsi_task *
+iscsi_prefetch10_task(struct iscsi_context *iscsi, int lun,
+		      uint32_t lba, int num_blocks,
+		      int immed, int group,
+		      iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_prefetch16_task(struct iscsi_context *iscsi, int lun,
+		      uint64_t lba, int num_blocks,
+		      int immed, int group,
+		      iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_read6_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       uint32_t datalen, int blocksize, iscsi_command_cb cb,
 		       void *private_data);
@@ -590,14 +609,64 @@ iscsi_read10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 EXTERN struct scsi_task *
 iscsi_write10_task(struct iscsi_context *iscsi, int lun,
 			unsigned char *data, uint32_t datalen, uint32_t lba, int fua,
-			int fuanv, int blocksize, iscsi_command_cb cb,
+			int fua_nv, int blocksize, iscsi_command_cb cb,
 			void *private_data);
 EXTERN struct scsi_task *
+iscsi_read12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_write12_task(struct iscsi_context *iscsi, int lun, uint32_t lba, 
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_read16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba, 
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_verify10_task(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint32_t lba,
+		    int vprotect, int dpo, int bytchk, 
+		    int blocksize, iscsi_command_cb cb,
+		    void *private_data);
+EXTERN struct scsi_task *
+iscsi_writesame10_task(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint32_t lba, uint16_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group,
+		       iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_writesame16_task(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint64_t lba, uint32_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group,
+		       iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_modesense6_task(struct iscsi_context *iscsi, int lun, int dbd,
 			   int pc, int page_code, int sub_page_code,
 			   unsigned char alloc_len, iscsi_command_cb cb,
 			   void *private_data);
 
+struct unmap_list {
+       uint64_t	  lba;
+       uint32_t	  num;
+};
+
+EXTERN struct scsi_task *
+iscsi_unmap_task(struct iscsi_context *iscsi, int lun, int anchor, int group,
+		 struct unmap_list *list, int list_len,
+		 iscsi_command_cb cb, void *private_data);
+
 
 /*
  * Sync commands for SCSI
@@ -625,19 +694,73 @@ EXTERN struct scsi_task *
 iscsi_read10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		  uint32_t datalen, int blocksize);
 
+EXTERN struct scsi_task *
+iscsi_write10_sync(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint32_t lba, int fua,
+			int fua_nv, int blocksize);
+
+EXTERN struct scsi_task *
+iscsi_read12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		  uint32_t datalen, int blocksize,
+		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_write12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_read16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		  uint32_t datalen, int blocksize,
+		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_write16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
+
 EXTERN struct scsi_task *
 iscsi_readcapacity10_sync(struct iscsi_context *iscsi, int lun, int lba,
 			  int pmi);
 
+EXTERN struct scsi_task *
+iscsi_readcapacity16_sync(struct iscsi_context *iscsi, int lun);
+
 EXTERN struct scsi_task *
 iscsi_synchronizecache10_sync(struct iscsi_context *iscsi, int lun, int lba,
 			      int num_blocks, int syncnv, int immed);
 
 EXTERN struct scsi_task *
-iscsi_write10_sync(struct iscsi_context *iscsi, int lun,
-			unsigned char *data, uint32_t datalen, uint32_t lba, int fua,
-			int fuanv, int blocksize);
+iscsi_prefetch10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		      int num_blocks, int immed, int group);
 
+EXTERN struct scsi_task *
+iscsi_prefetch16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		      int num_blocks, int immed, int group);
+
+EXTERN struct scsi_task *
+iscsi_verify10_sync(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint32_t lba,
+		    int vprotect, int dpo, int bytchk, 
+		    int blocksize);
+
+EXTERN struct scsi_task *
+iscsi_writesame10_sync(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint32_t lba, uint16_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group);
+
+EXTERN struct scsi_task *
+iscsi_writesame16_sync(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint64_t lba, uint32_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group);
+
+EXTERN struct scsi_task *
+iscsi_unmap_sync(struct iscsi_context *iscsi, int lun, int anchor, int group,
+		 struct unmap_list *list, int list_len);
 
 /*
  * This function is used when  the application wants to specify its own buffers to read the data
@@ -663,10 +786,9 @@ iscsi_write10_sync(struct iscsi_context *iscsi, int lun,
 EXTERN int scsi_task_add_data_in_buffer(struct scsi_task *task, int len, unsigned char *buf);
 
 /*
- * This function is used when you want to cancel a iscsi task.
- * The task will be immeidately cancelled and the callback for the task will
- * never be invoked.
- * The cancellation is only local in libiscsi. If the tast is already in-flight
+ * This function is used when you want to cancel a scsi task.
+ * The callback for the task will immediately be invoked with SCSI_STATUS_CANCELLED.
+ * The cancellation is only local in libiscsi. If the task is already in-flight
  * this call will not cancel the task at the target.
  * To cancel the task also a the target you need to call the task management functions.
  */
@@ -674,4 +796,14 @@ EXTERN int
 iscsi_scsi_task_cancel(struct iscsi_context *iscsi,
 		  struct scsi_task *task);
 
+/*
+ * This function is used when you want to cancel all scsi tasks.
+ * The callback for the tasks will immediately be invoked with SCSI_STATUS_CANCELLED.
+ * The cancellation is only local in libiscsi. If the tasks are already in-flight
+ * this call will not cancel the tasks at the target.
+ * To cancel the tasks also a the target you need to call the task management functions.
+ */
+EXTERN void
+iscsi_scsi_cancel_all_tasks(struct iscsi_context *iscsi);
+
 #endif /* __iscsi_h__ */
