@@ -324,7 +324,7 @@ int
 iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 			 struct iscsi_in_pdu *in)
 {
-	int statsn, flags, response, status;
+	int statsn, flags, status;
 	struct iscsi_scsi_cbdata *scsi_cbdata = pdu->scsi_cbdata;
 	struct scsi_task *task = scsi_cbdata->task;
 
@@ -349,8 +349,6 @@ iscsi_process_scsi_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		return -1;
 	}
 
-	response = in->hdr[2];
-
 	status = in->hdr[3];
 
 	switch (status) {
@@ -606,6 +604,27 @@ iscsi_readcapacity10_task(struct iscsi_context *iscsi, int lun, int lba,
 	return task;
 }
 
+struct scsi_task *
+iscsi_readcapacity16_task(struct iscsi_context *iscsi, int lun,
+			   iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+
+	task = scsi_cdb_readcapacity16();
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"readcapacity16 cdb.");
+		return NULL;
+	}
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
 struct scsi_task *
 iscsi_read6_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   uint32_t datalen, int blocksize,
@@ -662,6 +681,66 @@ iscsi_read10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 	return task;
 }
 
+struct scsi_task *
+iscsi_read12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of "
+				"the blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_read12(lba, datalen, blocksize, rdprotect,
+				dpo, fua, fua_nv, group_number);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"read12 cdb.");
+		return NULL;
+	}
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_read16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of "
+				"the blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_read16(lba, datalen, blocksize, rdprotect,
+				dpo, fua, fua_nv, group_number);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"read16 cdb.");
+		return NULL;
+	}
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
 struct scsi_task *
 iscsi_write10_task(struct iscsi_context *iscsi, int lun, unsigned char *data,
 		    uint32_t datalen, uint32_t lba, int fua, int fuanv, int blocksize,
@@ -679,7 +758,110 @@ iscsi_write10_task(struct iscsi_context *iscsi, int lun, unsigned char *data,
 	task = scsi_cdb_write10(lba, datalen, fua, fuanv, blocksize);
 	if (task == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
-				"read10 cdb.");
+				"write10 cdb.");
+		return NULL;
+	}
+
+	outdata.data = data;
+	outdata.size = datalen;
+
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_write12_task(struct iscsi_context *iscsi, int lun, uint32_t lba, 
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of the "
+				"blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_write12(lba, datalen, blocksize, wrprotect,
+				dpo, fua, fua_nv, group_number);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"write12 cdb.");
+		return NULL;
+	}
+
+	outdata.data = data;
+	outdata.size = datalen;
+
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba, 
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of the "
+				"blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_write16(lba, datalen, blocksize, wrprotect,
+				dpo, fua, fua_nv, group_number);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"write16 cdb.");
+		return NULL;
+	}
+
+	outdata.data = data;
+	outdata.size = datalen;
+
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_verify10_task(struct iscsi_context *iscsi, int lun, unsigned char *data,
+		    uint32_t datalen, uint32_t lba, int vprotect, int dpo, int bytchk, int blocksize,
+		    iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+
+	if (datalen % blocksize != 0) {
+		iscsi_set_error(iscsi, "Datalen:%d is not a multiple of the "
+				"blocksize:%d.", datalen, blocksize);
+		return NULL;
+	}
+
+	task = scsi_cdb_verify10(lba, datalen, vprotect, dpo, bytchk, blocksize);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"verify10 cdb.");
 		return NULL;
 	}
 
@@ -742,18 +924,192 @@ iscsi_synchronizecache10_task(struct iscsi_context *iscsi, int lun, int lba,
 	return task;
 }
 
+struct scsi_task *
+iscsi_prefetch10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		      int num_blocks, int immed, int group,
+		      iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+
+	task = scsi_cdb_prefetch10(lba, num_blocks, immed, group);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"prefetch10 cdb.");
+		return NULL;
+	}
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_prefetch16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		      int num_blocks, int immed, int group,
+		      iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+
+	task = scsi_cdb_prefetch16(lba, num_blocks, immed, group);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"prefetch16 cdb.");
+		return NULL;
+	}
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_writesame10_task(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint32_t lba, uint16_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group,
+		       iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+
+	task = scsi_cdb_writesame10(wrprotect, anchor, unmap, pbdata, lbdata, lba, group, num_blocks);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"writesame10 cdb.");
+		return NULL;
+	}
+
+	if (datalen) {
+		outdata.data = data;
+		outdata.size = datalen;
+		task->expxferlen = datalen;
+
+		if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+			scsi_free_scsi_task(task);
+			return NULL;
+		}
+	} else {
+		task->expxferlen = 0;
+		task->xfer_dir = SCSI_XFER_NONE;
+
+		if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+			scsi_free_scsi_task(task);
+			return NULL;
+		}
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_writesame16_task(struct iscsi_context *iscsi, int lun,
+		       unsigned char *data, uint32_t datalen,
+		       uint64_t lba, uint32_t num_blocks,
+		       int anchor, int unmap, int pbdata, int lbdata,
+		       int wrprotect, int group,
+		       iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+
+	task = scsi_cdb_writesame16(wrprotect, anchor, unmap, pbdata, lbdata, lba, group, num_blocks);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"writesame16 cdb.");
+		return NULL;
+	}
+
+	if (datalen) {
+		outdata.data = data;
+		outdata.size = datalen;
+		task->expxferlen = datalen;
+
+		if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+			scsi_free_scsi_task(task);
+			return NULL;
+		}
+	} else {
+		task->expxferlen = 0;
+		task->xfer_dir = SCSI_XFER_NONE;
+
+		if (iscsi_scsi_command_async(iscsi, lun, task, cb, NULL,
+				       private_data) != 0) {
+			scsi_free_scsi_task(task);
+			return NULL;
+		}
+	}
+
+	return task;
+}
+
+struct scsi_task *
+iscsi_unmap_task(struct iscsi_context *iscsi, int lun, int anchor, int group,
+		 struct unmap_list *list, int list_len,
+		 iscsi_command_cb cb, void *private_data)
+{
+	struct scsi_task *task;
+	struct iscsi_data outdata;
+	unsigned char *data;
+	int xferlen;
+	int i;
+
+	xferlen = 8 + list_len * 16;
+
+	task = scsi_cdb_unmap(anchor, group, xferlen);
+	if (task == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"unmap cdb.");
+		return NULL;
+	}
+
+	data = scsi_malloc(task, xferlen);
+	if (data == NULL) {
+		iscsi_set_error(iscsi, "Out-of-memory: Failed to create "
+				"unmap parameters.");
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+	*((uint16_t *)&data[0]) = htons(xferlen - 2);
+	*((uint16_t *)&data[2]) = htons(xferlen - 8);
+	for (i = 0; i < list_len; i++) {
+		*((uint32_t *)&data[8 + 16 * i])     = htonl(list[0].lba >> 32);
+		*((uint32_t *)&data[8 + 16 * i + 4]) = htonl(list[0].lba & 0xffffffff);
+		*((uint32_t *)&data[8 + 16 * i + 8]) = htonl(list[0].num);
+	}
+
+	outdata.data = data;
+	outdata.size = xferlen;
+
+	if (iscsi_scsi_command_async(iscsi, lun, task, cb, &outdata,
+				       private_data) != 0) {
+		scsi_free_scsi_task(task);
+		return NULL;
+	}
+
+	return task;
+}
+
 unsigned char *
 iscsi_get_user_in_buffer(struct iscsi_context *iscsi, struct iscsi_in_pdu *in, uint32_t pos, ssize_t *count)
 {
 	struct iscsi_pdu *pdu;
-	uint32_t len, offset;
+	uint32_t offset;
 	uint32_t itt;
 
 	if ((in->hdr[0] & 0x3f) != ISCSI_PDU_DATA_IN) {
 		return NULL;
 	}
 
-	len    = ntohl(*(uint32_t *)&in->hdr[4])&0x00ffffff;
 	offset = ntohl(*(uint32_t *)&in->hdr[40]);
 
 	itt = ntohl(*(uint32_t *)&in->hdr[16]);
@@ -768,3 +1124,9 @@ iscsi_get_user_in_buffer(struct iscsi_context *iscsi, struct iscsi_in_pdu *in, u
 
 	return scsi_task_get_data_in_buffer(pdu->scsi_cbdata->task, offset + pos, count);
 }
+
+struct scsi_task *
+iscsi_scsi_get_task_from_pdu(struct iscsi_pdu *pdu)
+{
+	return pdu->scsi_cbdata->task;
+}
