@@ -55,7 +55,7 @@ scsi_free_scsi_task(struct scsi_task *task)
 	free(task);
 }
 
-static void *
+void *
 scsi_malloc(struct scsi_task *task, size_t size)
 {
 	struct scsi_allocated_memory *mem;
@@ -95,6 +95,8 @@ const char *
 scsi_sense_key_str(int key)
 {
 	struct value_string keys[] = {
+		{SCSI_SENSE_HARDWARE_ERROR,
+		 "HARDWARE_ERROR"},
 		{SCSI_SENSE_ILLEGAL_REQUEST,
 		 "ILLEGAL_REQUEST"},
 		{SCSI_SENSE_UNIT_ATTENTION,
@@ -119,6 +121,8 @@ scsi_sense_ascq_str(int ascq)
 		 "LOGICAL_UNIT_NOT_SUPPORTED"},
 		{SCSI_SENSE_ASCQ_BUS_RESET,
 		 "BUS_RESET"},
+		{SCSI_SENSE_ASCQ_INTERNAL_TARGET_FAILURE,
+		 "INTERNAL_TARGET_FAILURE"},
 	       {0, NULL}
 	};
 
@@ -258,6 +262,36 @@ scsi_cdb_readcapacity10(int lba, int pmi)
 	return task;
 }
 
+/*
+ * service_action_in unmarshall
+ */
+static void *
+scsi_serviceactionin_datain_unmarshall(struct scsi_task *task)
+{
+	struct scsi_readcapacity16 *rc16;
+
+	switch (task->params.serviceactionin.sa) {
+	case SCSI_READCAPACITY16:
+		rc16 = scsi_malloc(task, sizeof(struct scsi_readcapacity16));
+		if (rc16 == NULL) {
+			return NULL;
+		}
+		rc16->returned_lba = htonl(*(uint32_t *)&(task->datain.data[0]));
+		rc16->returned_lba = (rc16->returned_lba << 32) | htonl(*(uint32_t *)&(task->datain.data[4]));
+		rc16->block_length = htonl(*(uint32_t *)&(task->datain.data[8]));
+		rc16->p_type       = (task->datain.data[12] >> 1) & 0x07;
+		rc16->prot_en      = task->datain.data[12] & 0x01;
+		rc16->p_i_exp      = (task->datain.data[13] >> 4) & 0x0f;
+		rc16->lbppbe       = task->datain.data[13] & 0x0f;
+		rc16->lbpme        = !!(task->datain.data[14] & 0x80);
+		rc16->lbprz        = !!(task->datain.data[14] & 0x40);
+		rc16->lalba        = htons(*(uint16_t *)&(task->datain.data[14])) & 0x3fff;
+		return rc16;
+	}
+
+	return NULL;
+}
+
 /*
  * parse the data in blob and calcualte the size of a full
  * readcapacity10 datain structure
@@ -341,13 +375,13 @@ scsi_inquiry_datain_getfullsize(struct scsi_task *task)
 
 	switch (task->params.inquiry.page_code) {
 	case SCSI_INQUIRY_PAGECODE_SUPPORTED_VPD_PAGES:
-		return task->datain.data[3] + 4;
+	case SCSI_INQUIRY_PAGECODE_BLOCK_DEVICE_CHARACTERISTICS:
 	case SCSI_INQUIRY_PAGECODE_UNIT_SERIAL_NUMBER:
 		return task->datain.data[3] + 4;
 	case SCSI_INQUIRY_PAGECODE_DEVICE_IDENTIFICATION:
-	     return ntohs(*(uint16_t *)&task->datain.data[2]) + 4;
-	case SCSI_INQUIRY_PAGECODE_BLOCK_DEVICE_CHARACTERISTICS:
-		return task->datain.data[3] + 4;
+	case SCSI_INQUIRY_PAGECODE_BLOCK_LIMITS:
+	case SCSI_INQUIRY_PAGECODE_LOGICAL_BLOCK_PROVISIONING:
+		return ntohs(*(uint16_t *)&task->datain.data[2]) + 4;
 	default:
 		return -1;
 	}
@@ -492,6 +526,34 @@ scsi_inquiry_datain_unmarshall(struct scsi_task *task)
 
 			dptr += dev->designator_length + 4;
 		}
+		return inq;
+	} else if (task->params.inquiry.page_code
+		   == SCSI_INQUIRY_PAGECODE_BLOCK_LIMITS) {
+		struct scsi_inquiry_block_limits *inq;
+
+		inq = scsi_malloc(task,
+		      sizeof(struct scsi_inquiry_block_limits));
+		if (inq == NULL) {
+			return NULL;
+		}
+		inq->periperal_qualifier   = (task->datain.data[0]>>5)&0x07;
+		inq->periperal_device_type = task->datain.data[0]&0x1f;
+		inq->pagecode              = task->datain.data[1];
+
+		inq->wsnz                  = task->datain.data[4] & 0x01;
+		inq->max_cmp               = task->datain.data[5];
+		inq->opt_gran              = ntohs(*(uint16_t *)&task->datain.data[6]);
+		inq->max_xfer_len          = ntohl(*(uint32_t *)&task->datain.data[8]);
+		inq->opt_xfer_len          = ntohl(*(uint32_t *)&task->datain.data[12]);
+		inq->max_prefetch          = ntohl(*(uint32_t *)&task->datain.data[16]);
+		inq->max_unmap             = ntohl(*(uint32_t *)&task->datain.data[20]);
+		inq->max_unmap_bdc         = ntohl(*(uint32_t *)&task->datain.data[24]);
+		inq->opt_unmap_gran        = ntohl(*(uint32_t *)&task->datain.data[28]);
+		inq->ugavalid              = !!(task->datain.data[32]&0x80);
+		inq->unmap_gran_align      = ntohl(*(uint32_t *)&task->datain.data[32]) & 0x7fffffff;
+		inq->max_ws_len            = ntohl(*(uint32_t *)&task->datain.data[36]);
+		inq->max_ws_len            = (inq->max_ws_len << 32) | ntohl(*(uint32_t *)&task->datain.data[40]);
+
 		return inq;
 	} else if (task->params.inquiry.page_code
 		   == SCSI_INQUIRY_PAGECODE_BLOCK_DEVICE_CHARACTERISTICS) {
@@ -506,8 +568,30 @@ scsi_inquiry_datain_unmarshall(struct scsi_task *task)
 		inq->periperal_device_type = task->datain.data[0]&0x1f;
 		inq->pagecode              = task->datain.data[1];
 
-		inq->medium_rotation_rate  = ntohs(*(uint16_t *)
-						   &task->datain.data[4]);
+		inq->medium_rotation_rate  = ntohs(*(uint16_t *)&task->datain.data[4]);
+		return inq;
+	} else if (task->params.inquiry.page_code
+		   == SCSI_INQUIRY_PAGECODE_LOGICAL_BLOCK_PROVISIONING) {
+		struct scsi_inquiry_logical_block_provisioning *inq;
+
+		inq = scsi_malloc(task,
+			   sizeof(struct scsi_inquiry_logical_block_provisioning));
+		if (inq == NULL) {
+			return NULL;
+		}
+		inq->periperal_qualifier   = (task->datain.data[0]>>5)&0x07;
+		inq->periperal_device_type = task->datain.data[0]&0x1f;
+		inq->pagecode              = task->datain.data[1];
+
+		inq->threshold_exponent = task->datain.data[4];
+		inq->lbpu               = !!(task->datain.data[5] & 0x80);
+		inq->lbpws              = !!(task->datain.data[5] & 0x40);
+		inq->lbpws10            = !!(task->datain.data[5] & 0x20);
+		inq->lbprz              = !!(task->datain.data[5] & 0x04);
+		inq->anc_sup            = !!(task->datain.data[5] & 0x02);
+		inq->dp	                = !!(task->datain.data[5] & 0x01);
+		inq->provisioning_type  = task->datain.data[6] & 0x07;
+
 		return inq;
 	}
 
@@ -591,11 +675,104 @@ scsi_cdb_read10(uint32_t lba, uint32_t xferlen, int blocksize)
 	return task;
 }
 
+/*
+ * READ12
+ */
+struct scsi_task *
+scsi_cdb_read12(uint32_t lba, uint32_t xferlen, int blocksize, int rdprotect, int dpo, int fua, int fua_nv, int group_number)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_READ12;
+
+	task->cdb[1] |= ((rdprotect & 0x07) << 5);
+	if (dpo) {
+		task->cdb[1] |= 0x10;
+	}
+	if (fua) {
+		task->cdb[1] |= 0x08;
+	}
+	if (fua_nv) {
+		task->cdb[1] |= 0x02;
+	}
+
+	*(uint32_t *)&task->cdb[2] = htonl(lba);
+	*(uint32_t *)&task->cdb[6] = htonl(xferlen/blocksize);
+
+	task->cdb[10] |= (group_number & 0x1f);
+
+	task->cdb_size = 12;
+	if (xferlen != 0) {
+		task->xfer_dir = SCSI_XFER_READ;
+	} else {
+		task->xfer_dir = SCSI_XFER_NONE;
+	}
+	task->expxferlen = xferlen;
+
+	task->params.read12.lba        = lba;
+	task->params.read12.num_blocks = xferlen/blocksize;
+
+	return task;
+}
+
+/*
+ * READ16
+ */
+struct scsi_task *
+scsi_cdb_read16(uint64_t lba, uint32_t xferlen, int blocksize, int rdprotect, int dpo, int fua, int fua_nv, int group_number)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_READ16;
+
+	task->cdb[1] |= ((rdprotect & 0x07) << 5);
+	if (dpo) {
+		task->cdb[1] |= 0x10;
+	}
+	if (fua) {
+		task->cdb[1] |= 0x08;
+	}
+	if (fua_nv) {
+		task->cdb[1] |= 0x02;
+	}
+
+	*(uint32_t *)&task->cdb[2] = htonl(lba >> 32);
+	*(uint32_t *)&task->cdb[6] = htonl(lba & 0xffffffff);
+	*(uint32_t *)&task->cdb[10] = htonl(xferlen/blocksize);
+
+	task->cdb[14] |= (group_number & 0x1f);
+
+	task->cdb_size = 16;
+	if (xferlen != 0) {
+		task->xfer_dir = SCSI_XFER_READ;
+	} else {
+		task->xfer_dir = SCSI_XFER_NONE;
+	}
+	task->expxferlen = xferlen;
+
+	task->params.read16.lba        = lba;
+	task->params.read16.num_blocks = xferlen/blocksize;
+
+	return task;
+}
+
 /*
  * WRITE10
  */
 struct scsi_task *
-scsi_cdb_write10(uint32_t lba, uint32_t xferlen, int fua, int fuanv, int blocksize)
+scsi_cdb_write10(uint32_t lba, uint32_t xferlen, int fua, int fua_nv, int blocksize)
 {
 	struct scsi_task *task;
 
@@ -610,7 +787,7 @@ scsi_cdb_write10(uint32_t lba, uint32_t xferlen, int fua, int fuanv, int blocksi
 	if (fua) {
 		task->cdb[1] |= 0x08;
 	}
-	if (fuanv) {
+	if (fua_nv) {
 		task->cdb[1] |= 0x02;
 	}
 
@@ -631,7 +808,264 @@ scsi_cdb_write10(uint32_t lba, uint32_t xferlen, int fua, int fuanv, int blocksi
 	return task;
 }
 
+/*
+ * WRITE12
+ */
+struct scsi_task *
+scsi_cdb_write12(uint32_t lba, uint32_t xferlen, int blocksize, int wrprotect, int dpo, int fua, int fua_nv, int group_number)
+{
+	struct scsi_task *task;
 
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_WRITE12;
+
+	task->cdb[1] |= ((wrprotect & 0x07) << 5);
+	if (dpo) {
+		task->cdb[1] |= 0x10;
+	}
+	if (fua) {
+		task->cdb[1] |= 0x08;
+	}
+	if (fua_nv) {
+		task->cdb[1] |= 0x02;
+	}
+
+	*(uint32_t *)&task->cdb[2] = htonl(lba);
+	*(uint32_t *)&task->cdb[6] = htonl(xferlen/blocksize);
+
+	task->cdb[10] |= (group_number & 0x1f);
+
+	task->cdb_size = 12;
+	if (xferlen != 0) {
+		task->xfer_dir = SCSI_XFER_WRITE;
+	} else {
+		task->xfer_dir = SCSI_XFER_NONE;
+	}
+	task->expxferlen = xferlen;
+
+	task->params.write12.lba        = lba;
+	task->params.write12.num_blocks = xferlen/blocksize;
+
+	return task;
+}
+
+/*
+ * WRITE16
+ */
+struct scsi_task *
+scsi_cdb_write16(uint64_t lba, uint32_t xferlen, int blocksize, int wrprotect, int dpo, int fua, int fua_nv, int group_number)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_WRITE16;
+
+	task->cdb[1] |= ((wrprotect & 0x07) << 5);
+	if (dpo) {
+		task->cdb[1] |= 0x10;
+	}
+	if (fua) {
+		task->cdb[1] |= 0x08;
+	}
+	if (fua_nv) {
+		task->cdb[1] |= 0x02;
+	}
+
+	*(uint32_t *)&task->cdb[2] = htonl(lba >> 32);
+	*(uint32_t *)&task->cdb[6] = htonl(lba & 0xffffffff);
+	*(uint32_t *)&task->cdb[10] = htonl(xferlen/blocksize);
+
+	task->cdb[14] |= (group_number & 0x1f);
+
+	task->cdb_size = 16;
+	if (xferlen != 0) {
+		task->xfer_dir = SCSI_XFER_WRITE;
+	} else {
+		task->xfer_dir = SCSI_XFER_NONE;
+	}
+	task->expxferlen = xferlen;
+
+	task->params.write16.lba        = lba;
+	task->params.write16.num_blocks = xferlen/blocksize;
+
+	return task;
+}
+
+/*
+ * VERIFY10
+ */
+struct scsi_task *
+scsi_cdb_verify10(uint32_t lba, uint32_t xferlen, int vprotect, int dpo, int bytchk, int blocksize)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_VERIFY10;
+
+	if (vprotect) {
+		task->cdb[1] |= ((vprotect << 5) & 0xe0);
+	}
+	if (dpo) {
+		task->cdb[1] |= 0x10;
+	}
+	if (bytchk) {
+		task->cdb[1] |= 0x02;
+	}
+
+	*(uint32_t *)&task->cdb[2] = htonl(lba);
+	*(uint16_t *)&task->cdb[7] = htons(xferlen/blocksize);
+
+	task->cdb_size = 10;
+	if (xferlen != 0) {
+		task->xfer_dir = SCSI_XFER_WRITE;
+	} else {
+		task->xfer_dir = SCSI_XFER_NONE;
+	}
+	task->expxferlen = xferlen;
+
+	task->params.verify10.lba        = lba;
+	task->params.verify10.num_blocks = xferlen/blocksize;
+	task->params.verify10.vprotect   = vprotect;
+	task->params.verify10.dpo        = dpo;
+	task->params.verify10.bytchk     = bytchk;
+
+	return task;
+}
+
+
+/*
+ * UNMAP
+ */
+struct scsi_task *
+scsi_cdb_unmap(int anchor, int group, uint16_t xferlen)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_UNMAP;
+
+	if (anchor) {
+		task->cdb[1] |= 0x01;
+	}
+	task->cdb[6] |= group & 0x1f;
+
+	*(uint16_t *)&task->cdb[7] = htons(xferlen);
+
+	task->cdb_size = 10;
+	task->xfer_dir = SCSI_XFER_WRITE;
+	task->expxferlen = xferlen;
+
+	return task;
+}
+
+/*
+ * WRITE_SAME10
+ */
+struct scsi_task *
+scsi_cdb_writesame10(int wrprotect, int anchor, int unmap, int pbdata, int lbdata, uint32_t lba, int group, uint16_t num_blocks)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_WRITE_SAME10;
+
+	if (wrprotect) {
+		task->cdb[1] |= ((wrprotect & 0x7) << 5);
+	}
+	if (anchor) {
+		task->cdb[1] |= 0x10;
+	}
+	if (unmap) {
+		task->cdb[1] |= 0x08;
+	}
+	if (pbdata) {
+		task->cdb[1] |= 0x04;
+	}
+	if (lbdata) {
+		task->cdb[1] |= 0x02;
+	}
+	*(uint32_t *)&task->cdb[2] = htonl(lba);
+	if (group) {
+		task->cdb[6] |= (group & 0x1f);
+	}
+	*(uint16_t *)&task->cdb[7] = htons(num_blocks);
+
+	task->cdb_size = 10;
+	task->xfer_dir = SCSI_XFER_WRITE;
+	task->expxferlen = 512;
+
+	return task;
+}
+
+/*
+ * WRITE_SAME16
+ */
+struct scsi_task *
+scsi_cdb_writesame16(int wrprotect, int anchor, int unmap, int pbdata, int lbdata, uint64_t lba, int group, uint32_t num_blocks)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_WRITE_SAME16;
+
+	if (wrprotect) {
+		task->cdb[1] |= ((wrprotect & 0x7) << 5);
+	}
+	if (anchor) {
+		task->cdb[1] |= 0x10;
+	}
+	if (unmap) {
+		task->cdb[1] |= 0x08;
+	}
+	if (pbdata) {
+		task->cdb[1] |= 0x04;
+	}
+	if (lbdata) {
+		task->cdb[1] |= 0x02;
+	}
+	*(uint32_t *)&task->cdb[2]  = htonl(lba >> 32);
+	*(uint32_t *)&task->cdb[6]  = htonl(lba & 0xffffffff);
+	*(uint32_t *)&task->cdb[10] = htonl(num_blocks);
+	if (group) {
+		task->cdb[14] |= (group & 0x1f);
+	}
+
+	task->cdb_size = 16;
+	task->xfer_dir = SCSI_XFER_WRITE;
+	task->expxferlen = 512;
+
+	return task;
+}
 
 /*
  * MODESENSE6
@@ -854,6 +1288,105 @@ scsi_cdb_synchronizecache10(int lba, int num_blocks, int syncnv, int immed)
 }
 
 
+/*
+ * PREFETCH10
+ */
+struct scsi_task *
+scsi_cdb_prefetch10(uint32_t lba, int num_blocks, int immed, int group)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_PREFETCH10;
+
+	if (immed) {
+		task->cdb[1] |= 0x02;
+	}
+	*(uint32_t *)&task->cdb[2] = htonl(lba);
+	task->cdb[6] |= group & 0x1f;
+	*(uint16_t *)&task->cdb[7] = htons(num_blocks);
+
+	task->cdb_size   = 10;
+	task->xfer_dir   = SCSI_XFER_NONE;
+	task->expxferlen = 0;
+
+	return task;
+}
+
+/*
+ * PREFETCH16
+ */
+struct scsi_task *
+scsi_cdb_prefetch16(uint64_t lba, int num_blocks, int immed, int group)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_PREFETCH16;
+
+	if (immed) {
+		task->cdb[1] |= 0x02;
+	}
+	*(uint32_t *)&task->cdb[2]  = htonl(lba >> 32);
+	*(uint32_t *)&task->cdb[6]  = htonl(lba & 0xffffffff);
+	*(uint32_t *)&task->cdb[10] = htonl(num_blocks);
+
+	task->cdb[14] |= group & 0x1f;
+
+	task->cdb_size   = 16;
+	task->xfer_dir   = SCSI_XFER_NONE;
+	task->expxferlen = 0;
+
+	return task;
+}
+
+/*
+ * SERVICEACTIONIN16
+ */
+struct scsi_task *
+scsi_cdb_serviceactionin16(enum scsi_service_action_in sa, uint32_t xferlen)
+{
+	struct scsi_task *task;
+
+	task = malloc(sizeof(struct scsi_task));
+	if (task == NULL) {
+		return NULL;
+	}
+
+	memset(task, 0, sizeof(struct scsi_task));
+	task->cdb[0]   = SCSI_OPCODE_SERVICE_ACTION_IN;
+
+	task->cdb[1] = sa;
+
+	*(uint32_t *)&task->cdb[10] = htonl(xferlen);
+
+	task->cdb_size   = 16;
+	task->xfer_dir   = SCSI_XFER_READ;
+	task->expxferlen = xferlen;
+
+	task->params.serviceactionin.sa = sa;
+
+	return task;
+}
+
+/*
+ * READCAPACITY16
+ */
+struct scsi_task *
+scsi_cdb_readcapacity16(void)
+{
+	return scsi_cdb_serviceactionin16(SCSI_READCAPACITY16, 32);
+}
 
 int
 scsi_datain_getfullsize(struct scsi_task *task)
@@ -891,6 +1424,8 @@ scsi_datain_unmarshall(struct scsi_task *task)
 		return NULL;
 	case SCSI_OPCODE_REPORTLUNS:
 		return scsi_reportluns_datain_unmarshall(task);
+	case SCSI_OPCODE_SERVICE_ACTION_IN:
+		return scsi_serviceactionin_datain_unmarshall(task);
 	}
 	return NULL;
 }
@@ -1131,7 +1666,7 @@ scsi_task_get_data_in_buffer(struct scsi_task *task, uint32_t pos, ssize_t *coun
 		}
 	}
 
-	if (count && *count > sdb->len - pos) {
+	if (count && *count > (ssize_t)(sdb->len - pos)) {
 		*count = sdb->len - pos;
 	}
 
@@ -1151,6 +1686,10 @@ iscsi_scsi_task_cancel(struct iscsi_context *iscsi,
 				SLIST_REMOVE(&task->in_buffers, ptr);
 			}
 			SLIST_REMOVE(&iscsi->waitpdu, pdu);
+			if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+				pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
+				      pdu->private_data);
+			}
 			iscsi_free_pdu(iscsi, pdu);
 			return 0;
 		}
@@ -1162,9 +1701,48 @@ iscsi_scsi_task_cancel(struct iscsi_context *iscsi,
 				SLIST_REMOVE(&task->in_buffers, ptr);
 			}
 			SLIST_REMOVE(&iscsi->outqueue, pdu);
+			if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+				pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
+				      pdu->private_data);
+			}
 			iscsi_free_pdu(iscsi, pdu);
 			return 0;
 		}
 	}
 	return -1;
 }
+
+void
+iscsi_scsi_cancel_all_tasks(struct iscsi_context *iscsi)
+{
+	struct iscsi_pdu *pdu;
+
+	for (pdu = iscsi->waitpdu; pdu; pdu = pdu->next) {
+		struct scsi_task *task = iscsi_scsi_get_task_from_pdu(pdu);
+	
+		while(task->in_buffers != NULL) {
+			struct scsi_data_buffer *ptr = task->in_buffers;
+			SLIST_REMOVE(&task->in_buffers, ptr);
+		}
+		SLIST_REMOVE(&iscsi->waitpdu, pdu);
+		if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+			pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
+				      pdu->private_data);
+		}
+		iscsi_free_pdu(iscsi, pdu);
+	}
+	for (pdu = iscsi->outqueue; pdu; pdu = pdu->next) {
+		struct scsi_task *task = iscsi_scsi_get_task_from_pdu(pdu);
+
+		while(task->in_buffers != NULL) {
+			struct scsi_data_buffer *ptr = task->in_buffers;
+			SLIST_REMOVE(&task->in_buffers, ptr);
+		}
+		SLIST_REMOVE(&iscsi->outqueue, pdu);
+		if ( !(pdu->flags & ISCSI_PDU_NO_CALLBACK)) {
+			pdu->callback(iscsi, SCSI_STATUS_CANCELLED, NULL,
+				      pdu->private_data);
+		}
+		iscsi_free_pdu(iscsi, pdu);
+	}
+}
